"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.JsonRpcError = exports.BrowserDependencies = exports.NotEthereumBrowserErrorMessage = void 0;
const ethereum_abi_encoder_1 = require("@zoltu/ethereum-abi-encoder");
const ethereum_crypto_1 = require("@zoltu/ethereum-crypto");
const ethereum_types_1 = require("@zoltu/ethereum-types");
exports.NotEthereumBrowserErrorMessage = `Your browser does not appear to be Ethereum enabled.`;
class BrowserDependencies {
    constructor(fallbackDependencies, options = {}) {
        this.fallbackDependencies = fallbackDependencies;
        this.options = options;
        this.call = async (to, methodSignature, parameters, value) => {
            if (!this.isEthereumBrowser() && this.fallbackDependencies !== undefined)
                return await this.fallbackDependencies.call(to, methodSignature, parameters, value);
            if (!this.isEthereumBrowser())
                throw new Error(exports.NotEthereumBrowserErrorMessage);
            const from = await this.getPrimaryAccount() || 0n;
            const data = await ethereum_abi_encoder_1.encodeMethod(ethereum_crypto_1.keccak256.hash, methodSignature, parameters);
            const gasPrice = (this.options.gasPriceInAttoethProvider !== undefined) ? await this.options.gasPriceInAttoethProvider() : await this.ethGasPrice();
            return await this.ethCall({ from, to, data, value, gasLimit: null, gasPrice });
        };
        this.submitTransaction = async (to, methodSignature, parameters, value) => {
            if (!this.isEthereumBrowser() && this.fallbackDependencies !== undefined)
                return await this.fallbackDependencies.submitTransaction(to, methodSignature, parameters, value);
            if (!this.isEthereumBrowser())
                throw new Error(exports.NotEthereumBrowserErrorMessage);
            const from = await this.getPrimaryAccount() || 0n;
            const data = await ethereum_abi_encoder_1.encodeMethod(ethereum_crypto_1.keccak256.hash, methodSignature, parameters);
            const gasPrice = (this.options.gasPriceInAttoethProvider !== undefined) ? await this.options.gasPriceInAttoethProvider() : await this.ethGasPrice();
            const gasEstimatingTransaction = { from, to, value, data, gasLimit: null, gasPrice };
            const gasLimit = this.options.gasLimitProvider !== undefined ? await this.options.gasLimitProvider(gasEstimatingTransaction) : await this.ethEstimateGas(gasEstimatingTransaction);
            const nonce = await this.ethGetTransactionCount(from);
            const unsignedTransaction = { ...gasEstimatingTransaction, gasLimit, nonce };
            const transactionHash = await this.ethSendTransaction(unsignedTransaction);
            let receipt = await this.ethGetTransactionReceipt(transactionHash);
            while (receipt === null || receipt.blockNumber === null) {
                await sleep(1000);
                receipt = await this.ethGetTransactionReceipt(transactionHash);
            }
            if (!receipt.status)
                throw new Error(`Transaction mined, but failed.`);
            return receipt;
        };
        this.request = async (method, params) => {
            if (window.ethereum === undefined)
                throw new Error(exports.NotEthereumBrowserErrorMessage);
            if ('request' in window.ethereum) {
                return await window.ethereum.request({ method, params });
            }
            else if ('sendAsync' in window.ethereum) {
                if (params !== undefined && !Array.isArray(params))
                    throw new Error(`Legacy Ethereum browsers do not support non-array RPC parameters.  ${JSON.stringify(params)}`);
                // we capture window.ethereum here to retain type narrowing since TS doesn't understand that the Promise callback function is executed immediately
                const ethereum = window.ethereum;
                return new Promise((resolve, reject) => {
                    ethereum.sendAsync({ jsonrpc: '2.0', id: 0, method, params: params || [] }, (error, response) => {
                        if (error)
                            return reject(unknownErrorToJsonRpcError(error, { request: { method, params } }));
                        if (!isJsonRpcLike(response))
                            return reject(new Error(exports.NotEthereumBrowserErrorMessage));
                        if ('error' in response && response.error !== null && response.error !== undefined)
                            return reject(unknownErrorToJsonRpcError(response.error, { request: { method, params } }));
                        if ('result' in response)
                            return resolve(response.result);
                        return reject(new Error(`Unexpected response from JSON-RPC: ${JSON.stringify(response)}`));
                    });
                });
            }
            else {
                throw new Error(exports.NotEthereumBrowserErrorMessage);
            }
        };
        this.isEthereumBrowser = () => {
            if (window.ethereum !== undefined)
                return true;
            if (window.web3 !== undefined)
                return true;
            return false;
        };
        this.getPrimaryAccount = async () => {
            const accounts = await this.ethAccounts();
            return (accounts.length === 0) ? undefined : accounts[0];
        };
        this.makeRequest = (Request, Response) => async (...args) => {
            const request = new Request(0, ...args);
            const rawRequest = request.wireEncode();
            const rawResponse = await this.request(rawRequest.method, rawRequest.params);
            const response = new Response({ jsonrpc: '2.0', id: rawRequest.id, result: rawResponse });
            return response.result;
        };
        this.ethAccounts = this.makeRequest(ethereum_types_1.Rpc.Eth.Accounts.Request, ethereum_types_1.Rpc.Eth.Accounts.Response);
        this.ethCall = this.makeRequest(ethereum_types_1.Rpc.Eth.Call.Request, ethereum_types_1.Rpc.Eth.Call.Response);
        this.ethEstimateGas = this.makeRequest(ethereum_types_1.Rpc.Eth.EstimateGas.Request, ethereum_types_1.Rpc.Eth.EstimateGas.Response);
        this.ethGasPrice = this.makeRequest(ethereum_types_1.Rpc.Eth.GasPrice.Request, ethereum_types_1.Rpc.Eth.GasPrice.Response);
        this.ethGetTransactionReceipt = this.makeRequest(ethereum_types_1.Rpc.Eth.GetTransactionReceipt.Request, ethereum_types_1.Rpc.Eth.GetTransactionReceipt.Response);
        this.ethGetTransactionCount = this.makeRequest(ethereum_types_1.Rpc.Eth.GetTransactionCount.Request, ethereum_types_1.Rpc.Eth.GetTransactionCount.Response);
        this.ethSendTransaction = this.makeRequest(ethereum_types_1.Rpc.Eth.SendTransaction.Request, ethereum_types_1.Rpc.Eth.SendTransaction.Response);
    }
}
exports.BrowserDependencies = BrowserDependencies;
/*
Helpers
 */
function sleep(delayInMilliseconds) {
    return new Promise(resolve => setTimeout(resolve, delayInMilliseconds));
}
function isJsonRpcLike(maybe) {
    return typeof maybe === 'object' && maybe !== null && ('result' in maybe || 'error' in maybe);
}
function mergeIn(target, source) {
    for (const key in source) {
        const targetValue = target[key];
        const sourceValue = source[key];
        if (targetValue === undefined || targetValue === null) {
            ;
            target[key] = sourceValue;
        }
        else if (isPlainObject(targetValue) && isPlainObject(sourceValue)) {
            mergeIn(targetValue, sourceValue);
        }
        else {
            // drop source[key], don't want to override the target value
        }
    }
    return target;
}
function isPlainObject(maybe) {
    if (typeof maybe !== 'object')
        return false;
    if (maybe === null)
        return false;
    if (Array.isArray(maybe))
        return false;
    // classes can get complicated so don't try to merge them.  What does it mean to merge two Promises or two Dates?
    if (Object.getPrototypeOf(maybe) !== Object.prototype)
        return false;
    return true;
}
function isJsonRpcErrorPayload(maybe) {
    if (typeof maybe !== 'object')
        return false;
    if (maybe === null)
        return false;
    return 'code' in maybe && 'message' in maybe;
}
function unknownErrorToJsonRpcError(error, extraData) {
    if (error instanceof Error) {
        const mutableError = error;
        mutableError.code = mutableError.code || -32603;
        mutableError.data = mutableError.data || extraData;
        if (isPlainObject(mutableError.data))
            mergeIn(mutableError.data, extraData);
        return error;
    }
    if (isJsonRpcErrorPayload(error)) {
        const data = isPlainObject(error.data) ? mergeIn(error.data, extraData) : error.data;
        return new JsonRpcError(error.code, error.message, data);
    }
    // if someone threw something besides an Error, wrap it up in an error
    return new JsonRpcError(-32603, `Unexpected thrown value.`, mergeIn({ error }, extraData));
}
class JsonRpcError extends Error {
    constructor(code, message, data) {
        super(message);
        this.code = code;
        this.data = data;
        this.name = this.constructor.name;
    }
}
exports.JsonRpcError = JsonRpcError;
//# sourceMappingURL=index.js.map